/* vim: set noexpandtab fileencoding=utf-8 nomodified nowrap textwidth=270 foldmethod=marker foldmarker={{{,}}} foldcolumn=4 ruler showcmd lcs=tab\:|- list: */
// ,g = gcc, exactly one space after "set"


#include "bios.h"
#include "bios_test.h"
#include "Arduino.h"


char state, x, y, waiting;
int score, high = 0;
char shape[10];       // holding the current shape (with offset)
const char minos[7][10] PROGMEM = { { 0, 0, 1, 0, 0, -1, 1, -1,   0, 1 },     // Quadrat (with SRS offset)
                                    { -1, -1, 0, -1, 0, 0, 1, 0,  0, 0 },     // Z
                                    { -1, 0, 0, 0, 0, -1, 1, -1,  0, 0 },     // neg. Z
                                    { -1, 0, 0, 0, 1, 0, 2, 0,    0, 1 },     // I
                                    { -1, 0, 0, 0, 1, 0, 0, -1,   0, 0 },     // Podest
                                    { -1, -1, -1, 0, 0, 0, 1, 0,  0, 0 },     // L
                                    { -1, 0, 0, 0, 1, 0, 1, -1,   0, 0 }  };  // neg. L

void drawintro()
{
  BIOS::clear();
  BIOS::text(F("A R D U I N O   B L O C K S"), 10, 6, VGA_CYAN);
  BIOS::text(F("Press  SPACE"), 23, 13, VGA_YELLOW);
}

void drawfield()
{
  BIOS::clear();
  for (byte i=0; i<5; i++) BIOS::text(F("<!..........!>"), i,12, VGA_CYAN);
  for (byte i=5; i<20; i++) BIOS::text(F("<!..........!>"),i, 12, VGA_WHITE);
  BIOS::text(F("<!==========!>"), 20, 12, VGA_CYAN);
  BIOS::text(F("  VVVVVVVVVV  "), 21, 12, VGA_MAGENTA);
  BIOS::text(F("SCORE 0"), 0, 0, VGA_GREEN);
  BIOS::text(F("HIGH"), 0, 27);    
  BIOS::set_cursor(0,32);
  BIOS::writeNum(high);
  BIOS::text(F("CONTROLS"),  3, 0, VGA_CYAN);
  BIOS::text(F("A - Left"), 5, 0, VGA_GRAY);
  BIOS::text(F("D - Right"),  6, 0, VGA_GRAY);
  BIOS::text(F("W - Rotate"), 7, 0, VGA_GRAY);
  BIOS::text(F("S - Drop"),  8, 0, VGA_GRAY);
}

boolean testShape(char x, char y)
{
  boolean isok = true;
  for (char i=0; i<8; i+=2)
  {
    if (x + shape[i] < 14+0 || x + shape[i] > 14+9) { isok = false; break; }         // linker / rechter Rand
    if (y + shape[i+1] < 0 || y + shape[i+1] > 19) { isok = false; break; }        // oberer / unterer Rand
    if (BIOS::vram[y + shape[i+1]][x + shape[i]] != '.') { isok = false; break; }   // Spielfeld
  }
  return isok;
}


void drawShape(char c)
{
  for (char i=0; i<8; i+=2) BIOS::vram[y + shape[i+1]][x + shape[i]] = c;        // Shape ins Feld schreiben  
}
void newShape()
{
  byte m = random(7);
  for (char i=0; i<10; i++) shape[i] = pgm_read_byte_near(&minos[m][i]);
  x = 14+4; y = 1;
  if (testShape(x, y)) drawShape('#');
  else
  {
    BIOS::text(F("GAME  OVER"), 10, 14, VGA_RED);
    BIOS::text(F("Press  SPACE"), 20, 13, VGA_YELLOW);
    state = 2;
  }
}

void rotShape()
{
  char rotshape[10];
  for(char i=0; i<10; i+=2)
  {
    rotshape[i] = shape[i+1];
    rotshape[i+1] = -shape[i];
  }
  for (char i=0; i<10; i++) shape[i] = rotshape[i];
  x += shape[8]; y+= shape[9];  // add rotated offset correction
}

void updateField()
{
  char anzrows = 0;                                                             // count the number of rows cleared
  for (char y=0; y<20; y++)                                                    // gehe alle Reihen durch
  {
    boolean rowfull = true;
    for(char x=14; x<14+10; x++) if (BIOS::vram[y][x] == '.') rowfull = false;            // ist diese Reihe voll?
    if (rowfull)
    {
      anzrows++;
      for(char i=y; i>0; i--) for(char x=14; x<14+10; x++) BIOS::vram[i][x] = BIOS::vram[i-1][x];   // kopiere ab dieser Reihe alles nach unten
      for(char x=14; x<14+10; x++) BIOS::vram[0][x] = '.';                                      // und lösche die obere Zeile
    }
  }
  switch (anzrows)
  {
    case 1: score += 40; break;
    case 2: score += 100; break;
    case 3: score += 200; break;
    case 4: score += 500; break;
  }
}

void tetris_setup()
{
BIOS::inverting=false;
  drawintro();
  state = 0;
}

void tetris_loop()
{ 
  int frame = BIOS::frames;
  byte key = BIOS::get_key();
  switch(state)
  {
    case 0:
      if (key == ' ')
      {
        randomSeed(BIOS::frames); random();
        drawfield();
        newShape();
        waiting = 40; BIOS::frames = 0; score = 0; state = 1;

      }
      break;
    case 1:
      if (key != 0)
      {
        if (key == 's') { waiting -= 40; score++; }
        else
        {
          drawShape('.');        // Shape ins Feld schreiben  
          switch (key)
          {
            case 'a':
              if (testShape(x-1, y)) x--;
              break;
            case 'd':
              if (testShape(x+1, y)) x++;
              break;
            case 'w':
              rotShape();                                               // drehe die Shape ccw
              if (testShape(x, y) == false)                             // prüfe, ob darunter KEIN Platz ist
              {
                if (testShape(x-1, y)) x--;                             // teste links
                else if (testShape(x+1, y)) x++;                        // teste rechts
                else if (testShape(x-2, y)) x-=2;                       // Sonderfall langer Balken
                else if (testShape(x+2, y)) x+=2;
                else { rotShape(); rotShape(); rotShape(); }            // geht nicht => Drehung rückgängig machen          
              }
              break;            
          }
          drawShape('#');        // Shape ins Feld schreiben  
        }
      }
      if (waiting-- < 0)                                          // FALLEN
      {
        waiting += (40 - (BIOS::frames >> 9));
        drawShape('.');
        if (testShape(x, y+1)) { y++; drawShape('#'); }
        else
        {
          drawShape('#');                   // fix shape
          updateField();                    // clear full rows
          BIOS::set_cursor(0, 6);BIOS::writeNum(score); 
          if (score > high) { high = score; BIOS::set_cursor(0,32);BIOS::writeNum(high); }
          newShape();                       // pick a new shape if possible
          drawShape('#');
        }
      }  
      break;
    case 2:
      if (key == ' ') { drawintro(); state = 0; }
      break;
  }
  while (frame == BIOS::frames);
}
void setup() {
BIOS::set_output(BIOS_RCA);      // Inicializace BIOS
//	BIOS::set_output();              // Inicializace BIOS
	BIOS::inverting=true;

/*
	//
//	BIOS::scroll();			// cursor not scrolled
//	BIOS::clear(' '+0x80,VGA_YELLOW);	// cursor to 0,0
	//
	int line=4;
	BIOS::set_cursor(line++, 10);    	// Nastavení kurzoru na (0, 0)
	BIOS::set_color(VGA_GREEN); 	// Nastavení barvy na zelenou
	BIOS::write('A');          	// Zapsání znaku 'A' na obrazovku
	//
//	BIOS::set_output(BIOS_RCA);      // Inicializace BIOS
	BIOS::set_cursor(line++, 5);    	// Nastavení kurzoru na (0, 0)
	BIOS::set_color(VGA_GREEN); 	// Nastavení barvy na zelenou
	BIOS::write('B');          	// Zapsání znaku 'A' na obrazovku
//	BIOS::set_output(BIOS_VGA);
	BIOS::invert();			// inverts character at actual cursor
	// print interprets \r\n\t chars, write doesnot
	BIOS::print('c');
	BIOS::print("Hello\r\n\t World");
	BIOS::print(F("from the \r\n\tFLASH"));
	BIOS::printNum(1234, 18,'#');
	BIOS::printHex(0x9876, 6);
	BIOS::printBin(0b011101101, 2);
	//
	BIOS::text("Write:",line++,1,VGA_DKRED);
	BIOS::write('C');
	BIOS::write("Hello\r\n\t World");
	BIOS::writeNum(1234);
	BIOS::writeHex(0x9867,8,'$');
	BIOS::writeBin(234);
	//
	BIOS::set_cursor(line++,6);
	BIOS::get_cursorY();
	BIOS::get_cursorX();
	//
	BIOS::set_rowcolor(line-3, VGA_RED);
	BIOS::set_color(VGA_GREEN);	// current row
	//
	BIOS::get_key();
	BIOS::get_scancode();
	BIOS::printBin(BIOS::ps2status,16,'#');
*/
//	while(1){};
};
extern "C" {extern void VRAMtest();};
extern "C" {extern void VRAMtest2();};
int counter=0;
uint16_t c;
void loop(){
	if( (c=BIOS::get_key()) ) BIOS::print(c);
	if (c=='\n') BIOS::print('\r');
	if (c==xF1) for (long i=0;i<10000;++i) BIOS::vram[((i/BIOS_COLS)%BIOS_ROWS)][i%BIOS_COLS]=uint8_t(i&0xFF);
	if (c==xF2) for (int i=0; i<1000;++i) VRAMtest();
	if (c==xF3) for (int i=0; i<1000;++i) VRAMtest2();
	if (c==xF4) { tetris_setup(); while(true) { tetris_loop();};};
	if (c==xF12){ if(BIOS::current_output==BIOS_VGA) { BIOS::set_output(BIOS_RCA); } else { BIOS::set_output(BIOS_VGA);} ; };
/*
 * 	counter++;
	BIOS::set_cursor(10,20);
	BIOS::writeNum(counter,10);
*/


};
